/*
*******************************************************************************
* Copyright (c) 1994 Chinchilla Software Limited
*******************************************************************************
* System    : Oracle date/time module
* Module    : ora_date.c
* $Source$
* Author(s) : Chris Dipple
* Date      : 20-JAN-94
* $Revision$
* Purpose   : Oracle /date time library
*
*******************************************************************************
* Change history
*
* Version     Date      Time      Done by               (Followed by comment)
* ----------  --------  --------  -------------
* $Log$
*
*******************************************************************************
*/

const char RCS_id[] = "$Header$";

#include <sys/types.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include <string.h>
#include <errno.h>
#include "ora_date.h"

/*
 * The implementation of date manipulation below
 *
 *   -  supports sub-set of ORACLE SQL functions
 *   -  does arithmetic on (time_t) seconds difference from 1970
 *      (for compatibility with time(2)). As time_t is currently
 *      implemented as a long (32 bit) signed integer these routines
 *      will not work after 2038.
 *   -  private; no external entry points provided
 */

typedef struct namestr
{
  short   n_c;			/* number of characters */
  char   *n_upper;		/* upper case representation */
  char   *n_lower;		/* lower case representation */
  char   *n_mixed;		/* mixed case representation */
} NAME;

/* holds the day names - got from message file by NL routines */
const NAME daynames[] = {
  { 6, "SUNDAY",    "sunday",    "Sunday"    },
  { 6, "MONDAY",    "monday",    "Monday"    },
  { 7, "TUESDAY",   "tuesday",   "Tuesday"   },
  { 9, "WEDNESDAY", "wednesday", "Wednesday" },
  { 8, "THURSDAY",  "thursday",  "Thursday"  },
  { 6, "FRIDAY",    "friday",    "Friday"    },
  { 8, "SATURDAY",  "saturday",  "Saturday"  }
};

/* holds the month names - got from message file by NL routines */
const NAME monthnames[] = {
  { 7, "JANUARY",   "january",   "January"   },
  { 8, "FEBRUARY",  "february",  "February"  },
  { 5, "MARCH",     "march",     "March"     },
  { 5, "APRIL",     "april",     "April"     },
  { 3, "MAY",       "may",       "May"       },
  { 4, "JUNE",      "june",      "June"      },
  { 4, "JULY",      "july",      "July"      },
  { 6, "AUGUST",    "august",    "August"    },
  { 9, "SEPTEMBER", "september", "September" },
  { 7, "OCTOBER",   "october",   "October"   },
  { 8, "NOVEMBER",  "november",  "November"  },
  { 8, "DECEMBER",  "december",  "December"  }
};

const short monthlens[] = {
  31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31
};

const short leaplens[] = {
  31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31
};

const short cum_monthlens[] = {
  0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334, 365
};

#define DO_DAY		(1 << 0)
#define DO_MONTH	(1 << 1)
#define DO_YEAR		(1 << 2)
#define DO_HOURS	(1 << 3)
#define DO_MINS		(1 << 4)
#define DO_SECS		(1 << 5)
#define DO_AMPM		(1 << 6)
#define DO_WKDAY	(1 << 7)
#define DO_SSM		(1 << 8)
#define DO_WEEK		(1 << 9)
#define DO_QUARTER	(1 << 10)
#define DO_CENTURY	(1 << 11)

/*
  Correct for the last couple of hundred years
*/
#define ISLEAPYEAR(y) ((((y) % 4) == 0 && ((y) % 100) != 0) || ((y) % 400) == 0)

#define STREQ(s1,s2,n)	!strncmp(s1, s2, n)
#define STRIEQ(s1,s2,n)	!strnicmp(s1, s2, n)

#define MINSECS		60L
#define HOURSECS	(60L * MINSECS)
#define DAYSECS		(HOURSECS * 24L)
#define WEEKSECS	(DAYSECS * 7L)
#define YEARSECS	(DAYSECS * 365L)
#define LYEARSECS	(DAYSECS * 366L)
#define JULIAN_OFFSET	2440588L

static time_t now;

/* Case insensitive compare */
static int strnicmp(const char *s1, const char *s2, unsigned n)
{
  for ( ; *s1 && n && toupper(*s1) == toupper(*s2); n--, s1++, s2++);

  return (n) ? ((toupper(*s1) > toupper(*s2)) ? 1 : -1) : 0;
}

/* validate seconds */
static short ss_val(char **date, time_t *secs)
{
  short   seconds;
  char   *s;

  if (*(s = *date) == '0')
    s++;
  if (!isdigit(*s))
  {
    *secs = (time_t) 0;
    *date = s;

    return -1;
  }
  seconds = *s - '0';
  if (s == *date && *s < '6' && isdigit(*(s + 1)))
    seconds = seconds * 10 + *++s - '0';
  *secs = (time_t) seconds;
  *date = ++s;

  return seconds;
}

/* validate seconds since midnight */
static long sssss_val(char **date, long *secs)
{
  long seconds;
  char *s;

  if (*(s = *date) == '0')
    s++;
  if (!isdigit(*s))
  {
    *secs = 0;
    *date = s;

    return -1;
  }
  for (seconds = 0; isdigit(*s); s++)
    seconds = seconds * 10 + *s - '0';
  *secs = (time_t) seconds;
  *date = s;

  return (seconds >= 0 && seconds <= DAYSECS) ? seconds : -1;
}

/* validate julian date */
static long j_val(char **date, time_t *julian)
{
  long days;
  char *s;

  if (*(s = *date) == '0')
    s++;
  if (!isdigit(*s))
  {
    *julian = 0;
    *date = s;

    return -1;
  }
  for (days = 0; isdigit(*s); s++)
    days = days * 10 + *s - '0';
  *julian = (time_t) ((days - JULIAN_OFFSET) * DAYSECS);
  *date = s;

  return 1;
}

/* validate week no */
static short week_val(char **date, int *wk)
{
  short week;
  char *s;

  if (*(s = *date) == '0')
    s++;
  if (!isdigit(*s))
  {
    *wk = 0;
    *date = s;

    return -1;
  }
  for (week = 0; isdigit(*s); s++)
    week = week * 10 + *s - '0';
  *wk = week;
  *date = s;

  return (week > 0 && week <= 53) ? week : -1;
}

/* validate minutes */
static short mi_val(char **date, time_t *secs)
{
  short   minutes;
  char   *s;

  if (*(s = *date) == '0')
    s++;
  if (!isdigit(*s))
  {
    *secs = (time_t) 0;
    *date = s;

    return -1;
  }
  minutes = *s - '0';
  if (s == *date && *s < '6' && isdigit(*(s + 1)))
    minutes = minutes * 10 + *++s - '0';
  *secs = minutes * 60;
  *date = ++s;

  return minutes;
}

/* validate (12) hours */
static short hh12_val(char **date, time_t *secs)
{
  short   hours;
  char   *s;

  if (*(s = *date) == '0')
    s++;
  if (!isdigit(*s))
  {
    *secs = (time_t) 0;
    *date = s;

    return 0;
  }
  hours = *s - '0';
  if (*s == '1' && s == *date)
  {
    if (*++s == '0')
      hours = 10;
    else if (*s == '1')
      hours = 11;
    else if (*s == '2')
      hours = 12;
    else
      s--;
  }
  *secs = ((time_t) hours) * HOURSECS;
  *date = ++s;

  return hours;
}

/* validate am/pm */
static short ampm_val(char **date, time_t *secs)
{
  char *s;

  if (*(s = *date) == 'A' || *s == 'a')
    *secs = (time_t) 0;
  else if (*s == 'P' || *s == 'p')
    *secs = HOURSECS * 12;
  else
  {
    *secs = (time_t) 0;
    *date = s;

    return 0;
  }
  s++;
  while (*s && !isalnum(*s))
    s++;
  if (*s != 'M' && *s != 'm')
  {
    *date = s;
    *secs = (time_t) 0;

    return 0;
  }
  *date = ++s;

  return 1;
}

/* validate 24-hour value */
static short hh24_val(char **date, time_t *secs)
{
  short   hours;
  char   *s;

  if (*(s = *date) == '0')
    s++;
  if (!isdigit(*s))
  {
    *date = s;
    *secs = (time_t) 0;

    return -1;
  }
  hours = *s - '0';
  if (s == *date)
  {
    if (*s == '1')
    {
      if (isdigit(*(s + 1)))
	hours = hours * 10 + *++s - '0';
    }
    else if (*s == '2')
    {
      if (*(s + 1) >= '0' && *(s + 1) <= '3')
	hours = hours * 10 + *++s - '0';
    }
  }
  *secs = ((time_t) (hours)) * HOURSECS;
  *date = ++s;

  return hours;
}

/* validate day number */
static short dd_val(char **date, time_t *secs)
{
  short   days;
  char   *s;

  if (*(s = *date) == '0')
    s++;
  if (!isdigit(*s))
  {
    *secs = (time_t) 0;
    *date = s;

    return 0;
  }
  days = *s - '0';
  if (s == *date)
  {
    if (*s < '3')
    {
      if (isdigit(*(s + 1)))
	days = days * 10 + *++s - '0';
    }
    else if (*s == '3')
    {
      if (*++s == '0')
	days = 30;
      else if (*s == '1')
	days = 31;
      else
	s--;
    }
  }
  *secs = ((time_t) (days - 1)) * DAYSECS;
  *date = ++s;

  return days;
}

/* validate month number */
static short mm_val(char **date, time_t *secs)
{
  short   months;
  char   *s;
  int     i;

  if (*(s = *date) == '0')
    s++;
  *secs = 0;
  if (!isdigit(*s))
  {
    *date = s;

    return 0;
  }
  months = *s - '0';
  if (*s == '1' && s == *date)
  {
    if (*++s == '0')
      months = 10;
    else if (*s == '1')
      months = 11;
    else if (*s == '2')
      months = 12;
    else
      s--;
  }
  /* if it's a leap year, this is adjusted later on */
  for (i = 0; i < months - 1; i++)
    *secs += DAYSECS * (long)monthlens[i];
  *date = ++s;

  return months;
}

/* validate month name - len > 0 if only len chars to be looked at */
static short month_val(char **date, time_t *secs, int len)
{
  short   months;
  int     i;

  for (months = 0; months != 12; months++)
    if (STRIEQ(monthnames[months].n_upper, *date,
	      (len) ? len : monthnames[months].n_c) ||
	STRIEQ(monthnames[months].n_mixed, *date,
	      (len) ? len : monthnames[months].n_c) ||
	STRIEQ(monthnames[months].n_lower, *date,
	      (len) ? len : monthnames[months].n_c))
      break;
  *secs = 0;
  if (months == 12) return 0;
  /* if it's a leap year, this is adjusted later on */
  for (i = 0; i != months; i++)
    *secs += DAYSECS * (long)monthlens[i];
  *date += (len) ? len : monthnames[months].n_c;

  return months + 1;
}

/* validate day name - len > 0 if only len chars to be looked at */
static short day_val(char **date, int len)
{
  short   wkday;

  for (wkday = 0; wkday < 7; wkday++)
    if (STRIEQ(daynames[wkday].n_upper, *date,
	      (len) ? len : daynames[wkday].n_c) ||
	STRIEQ(daynames[wkday].n_mixed, *date,
	      (len) ? len : daynames[wkday].n_c) ||
	STRIEQ(daynames[wkday].n_lower, *date,
	      (len) ? len : daynames[wkday].n_c))
      break;
  if (wkday == 7) return 0;
  *date += (len) ? len : daynames[wkday].n_c;

  return wkday + 1;
}

/* validate 2 character year number */
static short yy_val(char **date, time_t *secs)
{
  short   years;
  char   *s;

  if (*(s = *date) == '0')
    s++;
  if (!isdigit(*s))
  {
    *secs = (time_t) 0;
    *date = s;

    return 0;
  }
  years = *s - '0';
  if (s == *date && isdigit(*(s + 1)))
    years = 10 * years + *++s - '0';
  *secs = YEARSECS * (years - 70);
  *secs += (time_t) ((short) ((years - 69) / 4)) * DAYSECS;
  *date = ++s;

  return years + 1900;
}

/* validate 4 character year number */
static short yyyy_val(char **date, time_t *secs)
{
  short   years;
  char   *s;
  char    buf[8];

  memcpy(buf, *date, 4);
  buf[4] = '\0';
  years = (short) strtol(buf, &s, 10);
  /*
     Only valid between 1970 and 2037 because time_t implemented as
     signed long!
  */
  if (s == buf || years > 2038 || years < 1970)
  {
    *secs = (time_t) 0;

    return 0;
  }
  *secs = YEARSECS * (years - 1970);
  *secs += (time_t) ((short) ((years - 1969) / 4)) * DAYSECS;
  *date += 4;

  return years;
}

/*
 * ORACLE-like date validation
 *
 *  - knows nothing about things like  'st'
 *  - less choosy (not case sensitive, ignores odd spaces and punctuation)
 *  - hidden option format string SYSDATE gives you the current system date
 * returns
 *   0 for something invalid
 *   1 for something valid
 */
struct tm *date_val(char *date, char *fmt, char **where_went, time_t *totsecs)
{
  time_t  secs;
  time_t  week_secs;
  time_t  julian;
  long    ssm;
  int     week;
  short   months;
  short   wkday;
  short   years;
  short   days;
  short   state;
  char   *cp;
  long    days_since_dot;
  long    cur_day;
  char    buf[128];
  struct tm *tim;

#define DONE(x)	(state & (x))
#define DO(s, word, len) *totsecs += (s); state |= (word); fmt += (len)

  state = months = wkday = years = days = 0;
  *totsecs = 0;
  (void) strcpy(buf, fmt);
  /*
    Make case insensitive
  */
  for (cp = date; *cp; cp++)
    if (islower(*cp))
      *cp = toupper(*cp);
  for (cp = fmt = buf; *cp; cp++)
    if (islower(*cp))
      *cp = toupper(*cp);
  if (!strnicmp(date, "SYSDATE", 7) || !strnicmp(date, "NOW", 3))
  {
    (void)time(&now);
    tim = localtime(&now);
    *totsecs = (time_t) now;

    return tim;
  }
  while (*fmt && !isalnum(*fmt))
    fmt++;
  while (*fmt)
  {
    while (*date && !isalnum(*date))
      date++;
    switch (*fmt)
    {
    case 'J' :
      if (!*(fmt + 1) && j_val(&date, &julian) != -1)
      {
        DO(julian, DO_DAY | DO_WEEK | DO_MONTH | DO_YEAR, 1);
        break;
      }
    case 'S' :
      if (!DONE(DO_SECS) && *(fmt + 1) == 'S' && !*(fmt + 2)  &&
	  ss_val(&date, &secs) != -1)
      {
	DO(secs, DO_SECS, 2);
	break;
      }
      if (!DONE(DO_SSM) && STREQ("SSSSS", fmt, 2)  &&
	  (ssm = sssss_val(&date, &ssm)) != -1)
      {
	DO(ssm, DO_SSM | DO_SECS | DO_MINS | DO_HOURS | DO_AMPM, 5);
	break;
      }
      return (struct tm *)NULL;;
    case 'M' :
      if (!DONE(DO_MONTH))
      {
	if (STREQ("MM", fmt, 2)    &&
	    (months = mm_val(&date, &secs)) != 0)
	{
	  DO(secs, DO_MONTH, 2);
	  break;
	}
	if (STREQ("MONTH", fmt, 5) &&
	    (months = month_val(&date, &secs, 0)) != 0)
	{
	  DO(secs, DO_MONTH, 5);
	  break;
	}
	if (STREQ("MON", fmt, 3)   &&
	    (months = month_val(&date, &secs, 3)) != 0)
	{
	  DO(secs, DO_MONTH, 3);
	  break;
	}
      }
      if (!DONE(DO_MINS) && STREQ("MI", fmt, 2)  &&
	  mi_val(&date, &secs) != -1)
      {
	DO(secs, DO_MINS, 2);
	break;
      }
      return (struct tm *)NULL;
    case 'H' :
      if (!DONE(DO_HOURS | DO_AMPM))
      {
	if (STREQ("HH24", fmt, 4) && 
	    hh24_val(&date, &secs) != -1)
	{
	  DO(secs, DO_HOURS | DO_AMPM, 4);
	  break;
	}
	if (!DONE(DO_HOURS))
	{
	  if (STREQ("HH", fmt, 2)   &&
	      hh12_val(&date, &secs) != -1)
	  {
	    DO(secs, DO_HOURS, 2);
	    break;
	  }
	  if (STREQ("HH12", fmt, 4) &&
	      hh12_val(&date, &secs) != -1)
	  {
	    DO(secs, DO_HOURS, 4);
	    break;
	  }
	}
      }
      return (struct tm *)NULL;
    case 'A' :
    case 'P' :
      if (!DONE(DO_AMPM) && *(fmt + 1) == 'M' &&
	  ampm_val(&date, &secs) != 0)
      {
	DO(secs, DO_AMPM, 2);
	break;
      }
      return (struct tm *)NULL;
    case 'D' :
      if (!DONE(DO_DAY) && STREQ("DD", fmt, 2)    &&
	  (days = dd_val(&date, &secs)) != 0)
      {
	DO(secs, DO_DAY, 2);
	break;
      }
      if (!DONE(DO_WKDAY))
      {
	if (STREQ("DAY", fmt, 3) &&
	    (wkday = day_val(&date, 0)) != 0)
	{
	  DO(0, DO_WKDAY, 3);
	  break;
	}
	if (STREQ("DY", fmt, 2)  &&
	    (wkday = day_val(&date, 3)) != 0)
	{
	  DO(0, DO_WKDAY, 2);
	  break;
	}
      }
      return (struct tm *)NULL;
    case 'W' :
    case 'I' :
      if (!DONE(DO_WEEK) && (STREQ("WW", fmt, 2) || STREQ("IW", fmt, 2)) &&
	  (week = week_val(&date, &week)) != 0)
      {
	DO(week, DO_WEEK, 2);
	break;
      }
      return (struct tm *)NULL;
    case 'Y' :
      if (!DONE(DO_YEAR))
      {
	if (STREQ("YYYY", fmt, 4) &&
	    (years = yyyy_val(&date, &secs)) != 0)
	{
	  DO(secs, DO_YEAR, 4);
	  break;
	}
	if (STREQ("YY", fmt, 2)   &&
	    (years = yy_val(&date, &secs)) != 0)
	{
	  DO(secs, DO_YEAR, 2);
	  break;
	}
      }
      return (struct tm *)NULL;
    default :
      /*
       * Invalid; either unrecognised format sub-string or wrong
       */
      *where_went = date;

      return (struct tm *)NULL;
    }
    while (*fmt && !isalnum(*fmt))
      fmt++;
  }
  *where_went = date;

  /* sanity checks */
  if (DONE(DO_YEAR | DO_MONTH | DO_DAY))
  {
    if (!DONE(DO_YEAR) || !DONE(DO_MONTH) || !DONE(DO_DAY))
      return (struct tm *)NULL;
    if (ISLEAPYEAR(years))
    {
      if (days > leaplens[months - 1])
	return (struct tm *)NULL;
      if (months > 2)
	*totsecs += DAYSECS;
    }
    else if (days > monthlens[months - 1])
      return (struct tm *)NULL;
  }

  /* Check the day of the week */
  if (DONE(DO_WKDAY))
  {
    if (!DONE(DO_YEAR | DO_MONTH | DO_DAY))
    {
      /* day of the week only */
      (void)time(&now);
      tim = gmtime(&now);

      week_secs = (time_t) now;
      week_secs = week_secs / DAYSECS * DAYSECS;
      days_since_dot = (int) (week_secs / DAYSECS);
      cur_day = (days_since_dot % 7) + 1;
      if (cur_day == wkday)
	week_secs += DAYSECS * 7;
      else if (cur_day < wkday)
	week_secs += DAYSECS * (wkday - cur_day);
      else
	week_secs += DAYSECS * (7 - wkday + cur_day);
      *totsecs += week_secs;
      state |= DO_YEAR;
    }
    else
    {
      days_since_dot = *totsecs / DAYSECS;
      cur_day = (days_since_dot % 7) + 1;
      if (cur_day != wkday)
	return (struct tm *)NULL;
    }
  }
  tim = gmtime(totsecs);

  return (DONE(DO_YEAR | DO_SSM)) ? tim : (struct tm *)NULL;
}

/*
 *
 * ORACLE-like date to character conversion
 *
 *  - knows nothing about 'nd', 'st'
 *  - copies across odd spaces and punctuation in the format string
 *  - always returns a pointer to static (so must be copied on return)
 *  - gives up in the event of garbage in the format, returning as much
 *    as it managed to convert
 *
 */
char *to_char(time_t totsecs, const char *fmt)
{
  static char ret_date[80];
  short   state;
  char   *op;
  struct tm *tim;
  int     lower;
  const char *old_fmt;

#undef DO
#define DO(word, in, out) state |= (word); fmt += (in); op += (out)

  if (!fmt || !*fmt)
    fmt = "DD-MON-YY";

  now = (time_t)totsecs;
  tim = gmtime(&now);
  op = ret_date;
  state = 0;
  while (*fmt && !isalnum(*fmt))
    *op++ = *fmt++;
  old_fmt = fmt;
  while (*fmt)
  {
    lower = 0;
    switch (*fmt)
    {
    case 'J' :
    case 'j' :
      (void) sprintf(op, "%ld",
	       (long)(totsecs / DAYSECS) + JULIAN_OFFSET);
      op += strlen(op);
      *op = 0;

      return ret_date;
    case 'm' :
      lower = 1;
    case 'M' :
      if (!DONE(DO_MINS) && STRIEQ("MI", fmt, 2))
      {
	(void) sprintf(op, "%2.2d", tim->tm_min);
	DO(DO_MINS, 2, 2);
	break;
      }
      if (!DONE(DO_MONTH))
      {
	int month = tim->tm_mon;

	if (STRIEQ("MM", fmt, 2))
	{
	  (void) sprintf(op, "%2.2d", month + 1);
	  DO(DO_MONTH, 2, 2);
	  break;
	}
	if (STRIEQ("MONTH", fmt, 5))
	{
	  (void) sprintf(op, "%s", (lower) ? monthnames[month].n_lower :
		       islower(*(fmt + 1)) ? monthnames[month].n_mixed :
					     monthnames[month].n_upper);
	  DO(DO_MONTH, 5, monthnames[month].n_c);
	  break;
	}
	if (STRIEQ("MON", fmt, 3))
	{
	  (void) sprintf(op, "%3.3s", (lower) ? monthnames[month].n_lower :
			  islower(*(fmt + 1)) ? monthnames[month].n_mixed :
			                        monthnames[month].n_upper);
	  DO(DO_MONTH, 3, 3);
	  break;
	}
      }
      break;
    case 'd' :
      lower = 1;
    case 'D' :
      if (!DONE(DO_DAY))
      {
	char *ch = (char *)fmt;

	if (*ch == 'D' || *ch == 'd')
        {
	  if (*++ch == 'D' || *ch == 'd')
	  {
	    if (*++ch == 'D' || *ch == 'd')
	    {
	      (void) sprintf(op, "%3.3d", tim->tm_yday);
	      DO(DO_DAY, 3, 3);
	    }
	    else
	    {
	      (void) sprintf(op, "%2.2d", tim->tm_mday);
	      DO(DO_DAY, 2, 2);
	    }
	    break;
	  }
	  else if (!isalnum(*ch))
	  {
	    (void) sprintf(op, "%1.1d", tim->tm_wday + 1);
	    DO(DO_DAY, 1, 1);
	    break;
	  }
        }
      }
      if (!DONE(DO_WKDAY))
      {
	int wday = tim->tm_wday;

	if (STRIEQ("DAY", fmt, 3))
	{
	  (void) sprintf(op, "%s", (lower) ? daynames[wday].n_lower :
		       islower(*(fmt + 1)) ? daynames[wday].n_mixed :
					     daynames[wday].n_upper);
	  DO(DO_WKDAY, 3, daynames[wday].n_c);
	  break;
	}
	if (STRIEQ("DY", fmt, 2))
	{
	  (void) sprintf(op, "%3.3s", (lower) ? daynames[wday].n_lower :
		          islower(*(fmt + 1)) ? daynames[wday].n_mixed :
					        daynames[wday].n_upper);
	  DO(DO_WKDAY, 2, 3);
	  break;
	}
      }
      break;
    case 'h' :
    case 'H' :
      if (!DONE(DO_HOURS))
      {
	int hour = tim->tm_hour;

	if (STRIEQ("HH24", fmt, 4))
	{
	  (void) sprintf(op, "%2.2d", hour);
	  DO(DO_HOURS, 4, 2);
	  break;
	}
	if (STRIEQ("HH12", fmt, 2))
	{
	  (void) sprintf(op, "%2.2d", (hour % 12 == 0) ? 12 : (hour % 12));
	  DO(DO_HOURS, 4, 2);
	  break;
	}
	if (STRIEQ("HH", fmt, 2))
	{
	  (void) sprintf(op, "%2.2d", (hour % 12 == 0) ? 12 : (hour % 12));
	  DO(DO_HOURS, 2, 2);
	  break;
	}
      }
      break;
    case 'a' :
    case 'p' :
      lower = 1;
    case 'A' :
    case 'P' :
      if (!DONE(DO_AMPM) && (STRIEQ("PM", fmt, 2) || STRIEQ("AM", fmt, 2)))
      {
	if (lower)
	  (void) sprintf(op, (tim->tm_hour > 11) ? "pm" : "am");
	else
	  (void) sprintf(op, (tim->tm_hour > 11) ? "PM" : "AM");
	DO(DO_AMPM, 2, 2);
        break;
      }
      break;
    case 's' :
    case 'S' :
      if (!DONE(DO_SSM) && STRIEQ("SSSSS", fmt, 5))
      {
	(void) sprintf(op, "%5.5ld",
	  tim->tm_hour * (HOURSECS) + tim->tm_min * 60L + tim->tm_sec);
	DO(DO_SSM, 5, 5);
	break;
      }
      if (!DONE(DO_SECS) && STRIEQ("SS", fmt, 2))
      {
	(void) sprintf(op, "%2.2d", tim->tm_sec);
	DO(DO_SECS, 2, 2);
        break;
      }
      break;
    case 'c' :
    case 'C' :
      if (!DONE(DO_CENTURY) && STRIEQ("CC", fmt, 2))
      {
        if (tim->tm_year >= 100)
          strcmp(op, "20");
        else
          strcmp(op, "19");
	DO(DO_CENTURY, 2, 2);
      }
      break;
    case 'q' :
    case 'Q' :
      if (!DONE(DO_QUARTER))
      {
        switch (tim->tm_mon)
        {
          case 0  : case 1  : case 2  :
            op[0] = '1'; break;
          case 3  : case 4  : case 5  :
            op[0] = '2'; break;
          case 6  : case 7  : case 8  :
            op[0] = '3'; break;
          case 9  : case 10 : case 11 :
            op[0] = '4'; break;
        }
        op[1] = '\0';
	DO(DO_QUARTER, 1, 1);
        break;
      }
      break;
    case 'w' :
    case 'W' :
    case 'i' :
    case 'I' :
      if (!DONE(DO_WEEK))
      {
        if (STRIEQ("WW", fmt, 2) || STRIEQ("IW", fmt, 2))
	{
	  int week, year, days, wd;

	  week = (tim->tm_yday - tim->tm_wday + 7) / 7;
	  if (week == 0)
	  {
	    year = tim->tm_year - 1 + 1900;
	    days = ISLEAPYEAR(year) ? 366 : 365;
	    wd   = tim->tm_wday - tim->tm_yday;
	    week = (days - wd + 7) / 7;
	  }
	  (void) sprintf(op, "%2.2d", week);
	  DO(DO_WEEK, 2, 2);
          break;
	}
        if (STRIEQ("W", fmt, 1))
        {
	  int week, mon, year, days, wd;

	  week = (tim->tm_mday - tim->tm_wday + 7) / 7;
	  if (week == 0)
	  {
	    mon = (tim->tm_mon) ? 11 : (tim->tm_mon - 1);
	    year = tim->tm_year - 1 + 1900;
            if (ISLEAPYEAR(year))
              days = leaplens[mon];
            else
              days = monthlens[mon];
	    wd   = tim->tm_wday - tim->tm_mday;
	    week = (days - wd + 7) / 7;
	  }
	  (void) sprintf(op, "%1.1d", week);
	  DO(DO_WEEK, 1, 1);
          break;
        }
      }
      if (*fmt != 'I' && *fmt != 'i')
        break;
    case 'y' :
    case 'Y' :
      fmt++;
      if (!DONE(DO_YEAR))
      {
	if (STRIEQ("YYY", fmt, 3))
	{
	  (void) sprintf(op, "%4.4d", tim->tm_year + 1900);
	  DO(DO_YEAR, 3, 4);
	  break;
	}
	if (STRIEQ("YY", fmt, 2))
	{
	  (void) sprintf(op, "%3.3d", (tim->tm_year + 1900) % 1000);
	  DO(DO_YEAR, 2, 3);
	  break;
	}
	if (STRIEQ("Y", fmt, 1))
	{
	  (void) sprintf(op, "%2.2d", tim->tm_year % 100);
	  DO(DO_YEAR, 1, 2);
	  break;
	}
	(void) sprintf(op, "%1.1d", tim->tm_year % 10);
	DO(DO_YEAR, 0, 1);
      }
      break;
    default :
      *op = 0;
      return ret_date;
    }
    while (*fmt && !isalnum(*fmt))
      *op++ = *fmt++;
    if (old_fmt == fmt)
    {
      *op = 0;
      return ret_date;
    }
    old_fmt = fmt;
  }
  *op = 0;

  return ret_date;
}

/*
  For UNIX we have to parse the environment variable TZ for daylight savings
  time.

  It appears that TIMEOFFSET is based on PDT from GMT.  PDT is GMT - 7 hours.
  Keep this in mind when looking at the following code.  /etc/TIMEZONE and
  TIMEZONE(F) are also useful references.

  Format of TZ environment variable could be as simple as:
  TZ='PST8'	Which means name of the zone is 'PST' which is 8 hours west of
  GMT with no daylight savings time.

  More complicated values are:
  TZ='PST8PDT' Means PST is 8 hours west, PDT is name for daylight time, which
  follows current US law for the date and time of change and is 1 hour less
  than standard time.

  TZ='PST8:00:00PDT7:00:00' Means the same as the previous example.

  TZ='PST8:00:00PDT7:00:00;117/2:00:00,299/2:00:00' Means same as previous
  except change over to daylight time occurs on 117th day (0 based -- Jan 1 is
  day 0) at 2:00:00 AM.  Change over to standard time occurs on 299th day at
  2:00:00 AM.  Note that for the southern hemisphere start of daylight time
  will be at a date greater than end of daylight time.  Also note that leap
  year is never counted. That is March 1 is always day 59 (0 based) or J60
  even on leap years.

  Start and end dates for daylight savings time have other variations:
  J117/2:00:00 would be a 1 based calender -- Jan 1 is day 1. W15.6 Means
  week 15 (1 <= n <= 53) day of week 6 (Sunday == 0 -- 0 <= d <= 6).
  Week 53 always means the last week containing day 'd' even if there are
  actually only 52 weeks containing day d.

  M4.5.6 Means day 6 (Sunday) of week 5 (1 <= n <= 5) month 4 (April).  Week
  5 of a month is always the last week containing day 'd', whether there are
  actually 4 or 5 weeks containing day d.
*/

/* First Sunday in April, last Sunday in October.*/
const char US_RULE[] = ";M4.1.0/2,M10.5.0";

/*
  Return the number of seconds different from GMT that we are.
*/
static time_t gmt_offset(const char *tz_ptr)
{
  time_t offset = 0;

  if (tz_ptr && *tz_ptr)
  {
    while (*tz_ptr && *++tz_ptr != '+' && *tz_ptr != '-' &&
          !isdigit(*tz_ptr));
    if (*tz_ptr)
    {
      int sign = 1;
  
      offset = atol(tz_ptr) * HOURSECS;
      if (offset < 0)
        sign = 0;
  
      while (*tz_ptr && (*tz_ptr == '+' || *tz_ptr == '-' || isdigit(*tz_ptr)))
        tz_ptr++;
      if (*tz_ptr == ':')
        tz_ptr++;
      if (*tz_ptr)
      {
        offset += sign * atol(tz_ptr) * 60;
        while (*tz_ptr &&
	       (*tz_ptr == '+' || *tz_ptr == '-' || isdigit(*tz_ptr)))
          tz_ptr++;
        if (*tz_ptr == ':')
          tz_ptr++;
        if (*tz_ptr)
      	  offset += sign * atol(tz_ptr);
      }
    }
  }

  return offset;
}

/*
  Return the number of seconds since the beginning of the year that this
  daylight savings time expression represents.
*/
static void dst_offset(const char *p, time_t t, long *lb, long *ub)
{
  int i;
  long yday;
  long first_jan;
  time_t n;
  					/* Day of week, 0..6 (Sun..Sat) */
  int days_off;

  while (*p && *p != ';' && *p != ',')
    p++;
  if (!*p)
    p = US_RULE;
  n = t / DAYSECS;			/* Days since start of 1970 */
  i = (int)(n / 365L) + 1;		/* First guess at year */
  do {
    i--;
    yday = (int)(n - (long)i * 365L) - ((i + 3) >> 2);
  } while (yday < 0);

  first_jan = (n - yday) * DAYSECS;
  days_off = (int)(((first_jan / DAYSECS) + 1L) % 7L);

  for (i = 0; i != 2; i++)
  {
    int month;
    long result = 0L;

    if (isdigit(*++p))	/* expression in 0 based days? */
    {
      result += atol(p); /* Number of days into the year. */
      result *= DAYSECS;
    }
    else
    {
      result = -1;
      switch(*p++)
      {
        case 'J':
          result += atol(p); /* Number of days into the year. */
          result *= DAYSECS;
          break;
        case 'W':
          result = atol(p) * 7 * DAYSECS;
          while (isdigit(*p))
            p++;
          if (*p == '.')
            result += atol(p) * DAYSECS;
          break;
        case 'M':
          month = atoi(p) - 1;
      
          /* Get day of week of 1st day of this month. */
          days_off = (days_off + cum_monthlens[month]) % 7;
          while (isdigit(*p))
            p++;
          if (*p++ == '.')
          {
            int week = atoi(p) - 1;
      
            while(isdigit(*p))
              p++;
            if (*p++ == '.')
            {
              int temp
	         ,day = atoi(p);
      
              temp = week * 7 + (7 + day - days_off) % 7;
              if (temp + 3 >= monthlens[month])
            	temp -= 4;
              result = (long)(cum_monthlens[month] + temp) * DAYSECS;
            }
          }
          break;
      }
    }
  
    if ((p = strchr(p, '/')) != 0)
    {
      result += atol(++p) * HOURSECS;
      while(isdigit(*p))
        p++;
      if (*p == ':')
      {
        result += atol(++p) * 60;
        while (isdigit(*p))
          p++;
        if (*p == ':')
          result += atol(++p);
      }
    }
  
    if (i == 0)
      *lb = result + first_jan;
    else
      *ub = result + first_jan;
  }
}

/*
 *
 * ORACLE-like character to date conversion
 *
 *  - knows nothing about 'nd', 'st'
 *  - copies across odd spaces and punctuation in the format string
 *  - always returns a pointer to static (so must be copied on return)
 *  - gives up in the event of garbage in the format, returning as much
 *    as it managed to convert
 *
 */
time_t to_date(const char date_str[], const char *ifmt)
{
  char *where_went;
  time_t n;

  if (!ifmt || !*ifmt)
    ifmt = "DD-MON-YY";

  if (date_val((char *)date_str, (char *)ifmt, &where_went, &n) ==
       (struct tm *)NULL)
    n = (time_t)0;

  return n;
}

/*
  Truncation function differs slightly from Oracle. There are no hard
  coded truncations. It expects a partial format string and will
  truncate to precision of format thing. This is generally more useful.
*/
time_t date_trunc(time_t t, const char *ifmt)
{
  char *where_went;
  char *ch;
  time_t n;

  if (!ifmt || !*ifmt)
    ifmt = "J";

  ch = to_char(t, ifmt);
  if (!ch || date_val(ch, (char *)ifmt, &where_went, &n) == (struct tm *)NULL)
    n = (time_t)0;

  return n;
}

/*
  Oracle add_months function for whole months only.
*/
time_t add_months(time_t t, const int n)
{
  int i, j, k = 0, year, day;
  const short *ylen;

  day  = atoi(to_char(t, "DD"));
  year = atoi(to_char(t, "YYYY"));
  if (ISLEAPYEAR(year))
    ylen = leaplens;
  else
    ylen = monthlens;
  j = atoi(to_char(t, "MM")) - 1;
  if (n >= 0)
  {
    for (i = 0; i != n; i++)
    {
      k = (i + j) % 12;
      if (i && !k)
      {
        year++;
        if (ISLEAPYEAR(year))
          ylen = leaplens;
        else
          ylen = monthlens;
      }
      t += ylen[k] * DAYSECS;
    }
    if (day != atoi(to_char(t, "DD")))
    {
      t -= ylen[k] * DAYSECS;
      k = (k + 1) % 12;
      t += ylen[k] * DAYSECS;
    }
  }
  else
  {
    j += (12 * 1000) - 1;
    for (i = 0; i != n; i--)
    {
      k = (i + j) % 12;
      if (k == 11)
      {
        year--;
        if (ISLEAPYEAR(year))
          ylen = leaplens;
        else
          ylen = monthlens;
      }
      t -= ylen[k] * DAYSECS;
    }
    if (day != atoi(to_char(t, "DD")))
    {
      t += ylen[k] * DAYSECS;
      k = k ? k - 1 : 11;
      t -= ylen[k] * DAYSECS;
    }
  }  

  return t;
}

/*
  Lookup table for new_time function when using 'built in formats'.
*/

typedef struct {
  const char *tz;
  const char *tz_str;
} TZ_Offset;

const TZ_Offset tz_offset[] = {{ "CET", "CET-1CET1;M3.4.0/1,M9.5.0/1" },
                               { "WET", "WET-1DST1;M3.4.0/1,M9.5.0/1" },
                               { "GMT", "GMT0BST1;M3.4.0/1,M10.4.0/1" },
                               { "AST", "AST3ADT"  },
                               { "ADT", "AST3ADT"  },
                               { "BST", "BST10BDT" },
                               { "BDT", "BST10BDT" },
                               { "CST", "CST5CDT"  },
                               { "CDT", "CST5CDT"  },
                               { "EST", "EST4EDT"  },
                               { "EDT", "EST4EDT"  },
                               { "HST", "HST9HDT"  },
                               { "HDT", "HST9HDT"  },
                               { "MST", "MST6MDT"  },
                               { "MDT", "MST6MDT"  },
                               { "NST", "NST3:30:00NST0" },
                               { "PST", "PST7PDT"  },
                               { "PDT", "PST7PDT"  },
                               { "YST", "YST8YDT"  },
                               { "YDT", "YST8YDT"  },
                               { (char *)NULL, (char *)NULL }};

/*
  New time function, it 'knows' about the same timezones as Oracle
  and also will accept a standard unix TZ string (see format comments
  above).
*/
time_t new_time(time_t t, const char *tz1, const char *tz2)
{
  int i, j;
  long lb, ub;

  for (i = j = 0; tz_offset[i].tz; i++, j++)
  {
    if (!strcmp(tz1, tz_offset[i].tz))
      tz1 = tz_offset[i].tz_str;
    if (!strcmp(tz2, tz_offset[j].tz))
      tz2 = tz_offset[j].tz_str;
  }

  t += gmt_offset(tz1);
  t -= gmt_offset(tz2);

  dst_offset(tz1, t, &lb, &ub);
  i = t >= lb && t <= ub;

  dst_offset(tz2, t, &lb, &ub);
  j = t >= lb && t <= ub;

  if ( i && !j)
    t -= HOURSECS;
  if (!i && j )
    t += HOURSECS;

  return t;
}

/*
  Oracle next_day function
*/
time_t next_day(time_t t, const char *day)
{
  int i, j;

  if (isdigit(day[0]))
  {
    if (day[0] >= '1' && day[0] <= '7')
      i = atoi(day) - 1;
    else
      return (time_t)0;
  }
  else
    for (i = 0; i != 7; i++)
      if (STRIEQ(daynames[i].n_upper, day, 3))
        break;

  j = atoi(to_char(t, "D")) - 1;

  if (i < j)
    t += (long)(7 + (i - j)) * DAYSECS;
  else
    t += (long)(i - j) * DAYSECS;

  return t;
}

/*
  Oracle sysdate function
*/
time_t sysdate(void)
{
  char *where_went;
  time_t n;

  if (date_val("SYSDATE", "", &where_went, &n) == (struct tm *)NULL)
    n = (time_t)0;

  return n;
}

/*
  Oracle last_day (of month) function
*/
time_t last_day(time_t t)
{
  int day;
  int mon;

  day = atoi(to_char(t, "DD"));
  mon = atoi(to_char(t, "MM")) - 1;
  if (mon == 1)
  {
    int year = atoi(to_char(t, "YYYY"));

    return t + (((ISLEAPYEAR(year)) ? leaplens[mon] : monthlens[mon]) - day) *
           DAYSECS;
  }
  else
    return t + (monthlens[mon] - day) * DAYSECS;
}

/*
  Oracle months_between() function
*/
int months_between(time_t t1, time_t t2)
{
  return (atoi(to_char(t2, "YYYY")) - atoi(to_char(t1, "YYYY"))) * 12 +
         (atoi(to_char(t2, "MM"))   - atoi(to_char(t1, "MM")));
}

#ifdef MAIN
int main(int argc, char *argv[])
{
  char *ch;
  char *where_went;
  time_t n;
  char *ifmt = (char *)NULL;
  char *ofmt = (char *)NULL;

  if (argc < 2)
	 return 1;
  if (argc >= 3)
	 ifmt = argv[2];
  if (argc >= 4)
	 ofmt = argv[3];

  if (!ifmt || !*ifmt)
	 ifmt = "DD-MON-YY";
  if (!ofmt || !*ofmt)
	 ofmt = "DD.MM.YYYY HH24:MI:SS";

  ch = "Not Valid";
  if (date_val(argv[1], ifmt, &where_went, &n) != (struct tm *)NULL)
	 ch = to_char(n, ofmt);

  printf("%s\n", ch);

  ch = to_char(new_time(n, "GMT", "CET"), ofmt);

  printf("%s\n", ch);

  return 0;
}
#endif /* MAIN */

